@@ -13,10 +13,13 @@ import os
 import sys
 import traceback
 import warnings
+from enum import Enum
 from typing import Any, Callable, ContextManager, Dict, Optional
 
+import black
 from prompt_toolkit.document import Document
 from prompt_toolkit.formatted_text import (
+    HTML,
     FormattedText,
     PygmentsTokens,
     StyleAndTextTuples,
@@ -24,19 +27,21 @@ from prompt_toolkit.formatted_text import (
     merge_formatted_text,
     to_formatted_text,
 )
-from prompt_toolkit.formatted_text.utils import (
-    fragment_list_to_text,
-    fragment_list_width,
-    split_lines,
-)
+from prompt_toolkit.formatted_text.utils import fragment_list_to_text, split_lines
+from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
 from prompt_toolkit.key_binding.vi_state import InputMode
 from prompt_toolkit.patch_stdout import patch_stdout as patch_stdout_context
-from prompt_toolkit.shortcuts import clear_title, print_formatted_text, set_title
-from prompt_toolkit.utils import DummyContext
+from prompt_toolkit.shortcuts import (
+    PromptSession,
+    clear_title,
+    print_formatted_text,
+    set_title,
+)
+from prompt_toolkit.styles import BaseStyle
+from prompt_toolkit.utils import DummyContext, get_cwidth
 from pygments.lexers import PythonLexer, PythonTracebackLexer
 from pygments.token import Token
 
-from .eventloop import inputhook
 from .python_input import PythonInput
 
 __all__ = ["PythonRepl", "enable_deprecation_warnings", "run_config", "embed"]
@@ -107,17 +112,20 @@ class PythonRepl(PythonInput):
                 # Abort - try again.
                 self.default_buffer.document = Document()
             else:
-                self._process_text(text)
+                await self._process_text(text)
 
         if self.terminal_title:
             clear_title()
 
-    def _process_text(self, line: str) -> None:
+    async def _process_text(self, line: str) -> None:
 
         if line and not line.isspace():
+            if self.insert_blank_line_after_input:
+                self.app.output.write("\n")
+
             try:
                 # Eval and print.
-                self._execute(line)
+                await self._execute(line)
             except KeyboardInterrupt as e:  # KeyboardInterrupt doesn't inherit from Exception.
                 self._handle_keyboard_interrupt(e)
             except Exception as e:
@@ -129,7 +137,7 @@ class PythonRepl(PythonInput):
             self.current_statement_index += 1
             self.signatures = []
 
-    def _execute(self, line: str) -> None:
+    async def _execute(self, line: str) -> None:
         """
         Evaluate the line and print the result.
         """
@@ -170,66 +178,167 @@ class PythonRepl(PythonInput):
                 locals["_"] = locals["_%i" % self.current_statement_index] = result
 
                 if result is not None:
-                    self.show_result(result)
+                    await self.show_result(result)
             # If not a valid `eval` expression, run using `exec` instead.
             except SyntaxError:
                 code = compile_with_flags(line, "exec")
                 exec(code, self.get_globals(), self.get_locals())
 
-    def show_result(self, result: object) -> None:
+    async def show_result(self, result: object) -> None:
         """
         Show __repr__ for an `eval` result.
         """
         out_prompt = to_formatted_text(self.get_output_prompt())
-        result_repr = to_formatted_text("%r\n" % (result,))
 
-        # If __pt_repr__ is present, take this. This can return
-        # prompt_toolkit formatted text.
+        # If the repr is valid Python code, use the Pygments lexer.
+        result_repr = repr(result)
+        try:
+            compile(result_repr, "", "eval")
+        except SyntaxError:
+            formatted_result_repr = to_formatted_text(result_repr)
+        else:
+            # Syntactically correct. Format with black and syntax highlight.
+            if self.enable_output_formatting:
+                result_repr = black.format_str(
+                    result_repr,
+                    mode=black.FileMode(line_length=self.app.output.get_size().columns),
+                )
+
+            formatted_result_repr = to_formatted_text(
+                PygmentsTokens(list(_lex_python_result(result_repr)))
+            )
+
+        # If __pt_repr__ is present, take this. This can return prompt_toolkit
+        # formatted text.
         if hasattr(result, "__pt_repr__"):
             try:
-                result_repr = to_formatted_text(getattr(result, "__pt_repr__")())
-                if isinstance(result_repr, list):
-                    result_repr = FormattedText(result_repr)
+                formatted_result_repr = to_formatted_text(
+                    getattr(result, "__pt_repr__")()
+                )
+                if isinstance(formatted_result_repr, list):
+                    formatted_result_repr = FormattedText(formatted_result_repr)
             except:
                 pass
 
-        # If we have a string so far, and it's valid Python code,
-        # use the Pygments lexer.
-        if isinstance(result, str):
-            try:
-                compile(result, "", "eval")
-            except SyntaxError:
-                pass
-            else:
-                result = PygmentsTokens(list(_lex_python_result(result)))
-
         # Align every line to the prompt.
         line_sep = "\n" + " " * fragment_list_width(out_prompt)
         indented_repr: StyleAndTextTuples = []
 
-        for fragment in split_lines(result_repr):
+        lines = list(split_lines(formatted_result_repr))
+
+        for i, fragment in enumerate(lines):
             indented_repr.extend(fragment)
-            indented_repr.append(("", line_sep))
-        if indented_repr:
-            indented_repr.pop()
-        indented_repr.append(("", "\n"))
+
+            # Add indentation separator between lines, not after the last line.
+            if i != len(lines) - 1:
+                indented_repr.append(("", line_sep))
 
         # Write output tokens.
         if self.enable_syntax_highlighting:
             formatted_output = merge_formatted_text([out_prompt, indented_repr])
         else:
             formatted_output = FormattedText(
-                out_prompt + [("", fragment_list_to_text(result_repr))]
+                out_prompt + [("", fragment_list_to_text(formatted_result_repr))]
             )
 
+        if self.enable_pager:
+            await self._print_paginated_formatted_text(
+                to_formatted_text(formatted_output)
+            )
+        else:
+            self.print_formatted_text(to_formatted_text(formatted_output))
+
+        self.app.output.flush()
+
+    def print_formatted_text(self, formatted_text: StyleAndTextTuples) -> None:
         print_formatted_text(
-            formatted_output,
+            FormattedText(formatted_text),
             style=self._current_style,
             style_transformation=self.style_transformation,
             include_default_pygments_style=False,
             output=self.app.output,
         )
-        self.app.output.flush()
+
+    async def _print_paginated_formatted_text(
+        self, formatted_text: StyleAndTextTuples
+    ) -> None:
+        """
+        Print formatted text, using --MORE-- style pagination.
+        (Avoid filling up the terminal's scrollback buffer.)
+        """
+        pager_prompt = self.create_pager_prompt()
+        size = self.app.output.get_size()
+
+        abort = False
+
+        # Max number of lines allowed in the buffer before painting.
+        max_rows = size.rows - 1
+
+        # Page buffer.
+        rows_in_buffer = 0
+        columns_in_buffer = 0
+        page: StyleAndTextTuples = []
+
+        def flush_page() -> None:
+            nonlocal page, columns_in_buffer, rows_in_buffer
+            self.print_formatted_text(page)
+            page = []
+            columns_in_buffer = 0
+            rows_in_buffer = 0
+
+        async def show_pager() -> None:
+            nonlocal abort, max_rows
+
+            continue_result = await pager_prompt.prompt_async()
+            if continue_result == PagerResult.ABORT:
+                print("...")
+                abort = True
+
+            elif continue_result == PagerResult.NEXT_LINE:
+                max_rows = 1
+
+            elif continue_result == PagerResult.NEXT_PAGE:
+                max_rows = size.rows - 1
+
+        # Loop over lines. Show --MORE-- prompt when page is filled.
+        for line in split_lines(formatted_text):
+            for style, text, *_ in line:
+                for c in text:
+                    width = get_cwidth(c)
+
+                    # (Soft) wrap line if it doesn't fit.
+                    if columns_in_buffer + width > size.columns:
+                        # Show pager first if we get too many lines after
+                        # wrapping.
+                        if rows_in_buffer + 1 >= max_rows:
+                            flush_page()
+                            await show_pager()
+                            if abort:
+                                return
+
+                        rows_in_buffer += 1
+                        columns_in_buffer = 0
+
+                    columns_in_buffer += width
+                    page.append((style, c))
+
+            if rows_in_buffer + 1 >= max_rows:
+                flush_page()
+                await show_pager()
+                if abort:
+                    return
+            else:
+                page.append(("", "\n"))
+                rows_in_buffer += 1
+                columns_in_buffer = 0
+
+        flush_page()
+
+    def create_pager_prompt(self) -> PromptSession["PagerResult"]:
+        """
+        Create pager --MORE-- prompt.
+        """
+        return create_pager_prompt(self._current_style)
 
     def _handle_exception(self, e: Exception) -> None:
         output = self.app.output
@@ -411,3 +520,54 @@ def embed(
     else:
         with patch_context:
             repl.run()
+
+
+class PagerResult(Enum):
+    ABORT = "ABORT"
+    NEXT_LINE = "NEXT_LINE"
+    NEXT_PAGE = "NEXT_PAGE"
+
+
+def create_pager_prompt(style: BaseStyle) -> PromptSession[PagerResult]:
+    """
+    Create a "continue" prompt for paginated output.
+    """
+    bindings = KeyBindings()
+
+    @bindings.add("enter")
+    @bindings.add("down")
+    def next_line(event: KeyPressEvent) -> None:
+        event.app.exit(result=PagerResult.NEXT_LINE)
+
+    @bindings.add("space")
+    def next_page(event: KeyPressEvent) -> None:
+        event.app.exit(result=PagerResult.NEXT_PAGE)
+
+    @bindings.add("q")
+    @bindings.add("c-c")
+    @bindings.add("c-d")
+    @bindings.add("escape", eager=True)
+    def no(event: KeyPressEvent) -> None:
+        event.app.exit(result=PagerResult.ABORT)
+
+    @bindings.add("<any>")
+    def _(event: KeyPressEvent) -> None:
+        " Disallow inserting other text. "
+        pass
+
+    style
+
+    session: PromptSession[PagerResult] = PromptSession(
+        HTML(
+            "<status-toolbar>"
+            "<more> -- MORE -- </more> "
+            "<key>[Enter]</key> Scroll "
+            "<key>[Space]</key> Next page "
+            "<key>[q]</key> Quit "
+            "</status-toolbar>: "
+        ),
+        key_bindings=bindings,
+        erase_when_done=True,
+        style=style,
+    )
+    return session
